// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/18AISS8AO8/18AISS8AO8_1.x_GSC_DN/driver/device.c $
// $Rev: 51963 $
// $Date: 2022-10-26 10:20:45 -0500 (Wed, 26 Oct 2022) $

// 18AISS8AO8: Device Driver: source file

#include "main.h"



// variables ******************************************************************

const gsc_dev_id_t	dev_id_list[]	=
{
	// model		Vendor	Device	SubVen	SubDev	type
	{ "18AISS8AO8",	0x10B5, 0x9056, 0x10B5, 0x3251,	GSC_DEV_TYPE_18AISS8AO8	},
	{ NULL }
};



//*****************************************************************************
static void _ai_underflow_compute(dev_data_t* dev)
{
	if ((dev->cache.gsc_acr_32 & 0xFFF) > 0x002)
	{
		dev->cache.ai_underflow	= 1;
	}
	else
	{
		dev->cache.ai_underflow	= 0;
	}
}



//*****************************************************************************
static void _channels_compute(dev_data_t* dev)
{
	if (dev->cache.gsc_acr_32 & D16)
	{
		dev->cache.ai_chan_qty	= 4;
		dev->cache.ao_chan_qty	= 4;
	}
	else
	{
		dev->cache.ai_chan_qty	= 8;
		dev->cache.ao_chan_qty	= 8;
	}

	dev->cache.ai_chans_max	= 8;
	dev->cache.ao_chans_max	= 8;
}



//*****************************************************************************
static void _dmdma_compute(dev_data_t* dev)
{
	if ((dev->cache.gsc_acr_32 & 0xFFF) > 0x002)
		dev->cache.dmdma	= 1;
}



//*****************************************************************************
static void _fifos_compute(dev_data_t* dev)
{
	if (dev->cache.gsc_acr_32 & D17)
	{
		dev->cache.fifo_size_rx		= _256K;
		dev->cache.fifo_size_tx		= _256K;
	}
	else
	{
		dev->cache.fifo_size_rx		= _512K;
		dev->cache.fifo_size_tx		= _512K;
	}
}



//*****************************************************************************
static void _filters_compute(dev_data_t* dev)
{
	switch (dev->cache.gsc_acr_32 & (D19 | D18))
	{
		default:

			dev->cache.filter_a	= -1;		// not known
			dev->cache.filter_b	= -1;		// not known
			break;

		case 0:

			dev->cache.filter_a	= 80000;
			dev->cache.filter_b	= 200000;
			break;
	}
}



//*****************************************************************************
static void _master_clock_compute(dev_data_t* dev)
{
	switch (dev->cache.gsc_acr_32 & (D21 | D20))
	{
		default:
		case 0:		dev->cache.master_clock	= 40320000L;	break;
	}
}



/******************************************************************************
*
*	Function:	dev_device_create
*
*	Purpose:
*
*		Do everything needed to setup and use the given device.
*
*	Arguments:
*
*		dev		The structure to initialize.
*
*	Returned:
*
*		0		All is well.
*		< 0		An appropriate error status.
*
******************************************************************************/

int dev_device_create(dev_data_t* dev)
{
	static const gsc_bar_maps_t	bar_map	=
	{
		{
			// mem	io	rw
			{ 1,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 0: PLX registers, memory mapped
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 1: PLX registers, I/O mapped
			{ 1,	0,	GSC_REG_TYPE_ACCESS_RW },	// BAR 2: GSC registers, memory mapped
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 3: unused
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 4: unused
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO }	// BAR 5: unused
		}
	};

	u32	dma0;
	u32	dma1;
	int	ret;

	for (;;)	// A convenience loop.
	{
		// Verify some macro contents.
		ret	= gsc_macro_test_base_name(AISS8AO8_BASE_NAME);
		if (ret)	break;

		ret	= gsc_macro_test_model();
		if (ret)	break;

		// PCI setup.
		ret	= os_pci_dev_enable(&dev->pci);
		if (ret)	break;

		ret	= os_pci_master_set(&dev->pci);
		if (ret)	break;

		// Control ISR access to the device and data structure.
		ret	= os_spinlock_create(&dev->spinlock);
		if (ret)	break;

		// Control access to the device and data structure.
		ret	= os_sem_create(&dev->sem);
		if (ret)	break;

		// Access the BAR regions.
		ret	= gsc_bar_create(dev, &dev->bar, &bar_map);
		if (ret)	break;

		// Firmware access.
		dev->vaddr.gsc_acr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AISS8AO8_GSC_ACR));
		dev->vaddr.gsc_bcr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AISS8AO8_GSC_BCR));
		dev->vaddr.gsc_boor_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AISS8AO8_GSC_BOOR));
		dev->vaddr.gsc_clsr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AISS8AO8_GSC_CLSR));
		dev->vaddr.gsc_diopr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AISS8AO8_GSC_DIOPR));
		dev->vaddr.gsc_ibbsr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AISS8AO8_GSC_IBBSR));
		dev->vaddr.gsc_ibdr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AISS8AO8_GSC_IBDR));
		dev->vaddr.gsc_ibsr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AISS8AO8_GSC_IBSR));
		dev->vaddr.gsc_ibtr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AISS8AO8_GSC_IBTR));
		dev->vaddr.gsc_icr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AISS8AO8_GSC_ICR));
		dev->vaddr.gsc_obdr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AISS8AO8_GSC_OBDR));
		dev->vaddr.gsc_obsr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AISS8AO8_GSC_OBSR));
		dev->vaddr.gsc_obtr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AISS8AO8_GSC_OBTR));
		dev->vaddr.gsc_ocr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AISS8AO8_GSC_OCR));
		dev->vaddr.gsc_psr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AISS8AO8_GSC_PSR));
		dev->vaddr.gsc_ragr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AISS8AO8_GSC_RAGR));
		dev->vaddr.gsc_rbgr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AISS8AO8_GSC_RBGR));
		dev->vaddr.gsc_rcgr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AISS8AO8_GSC_RCGR));
		dev->vaddr.gsc_rdgr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AISS8AO8_GSC_RDGR));

		// Data cache initialization.
		dev->cache.gsc_acr_32		= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_acr_32);

		_ai_underflow_compute(dev);
		_channels_compute(dev);
		_dmdma_compute(dev);
		_fifos_compute(dev);
		_filters_compute(dev);
		_master_clock_compute(dev);

		dev->cache.auto_cal_ms		= 15000;
		dev->cache.initialize_ms	= 3;
		dev->cache.ndiv_max			= 0xFFFFFF;
		dev->cache.ndiv_min			= dev->cache.master_clock / 500000;

		dev->cache.fsamp_max		= dev->cache.master_clock / dev->cache.ndiv_min;
		dev->cache.fsamp_min		= dev->cache.master_clock / dev->cache.ndiv_max;

		// Initialize additional resources.
		ret	= dev_irq_create(dev);
		if (ret)	break;

		ret	= dev_io_create(dev);
		if (ret)	break;

		dma0	= GSC_DMA_SEL_STATIC
				| GSC_DMA_CAP_BMDMA_READ;
		dma1	= GSC_DMA_SEL_STATIC
				| GSC_DMA_CAP_BMDMA_WRITE;

		if (dev->cache.dmdma)
		{
			dma0	|= GSC_DMA_CAP_DMDMA_READ;
			dma1	|= GSC_DMA_CAP_DMDMA_WRITE;
		}

		ret		= gsc_dma_create(dev, dma0, dma1);
		break;
	}

	return(ret);
}



/******************************************************************************
*
*	Function:	dev_device_destroy
*
*	Purpose:
*
*		Do everything needed to release the referenced device.
*
*	Arguments:
*
*		dev		The partial data for the device of interest.
*
*	Returned:
*
*		None.
*
******************************************************************************/

void dev_device_destroy(dev_data_t* dev)
{
	if (dev)
	{
		gsc_dma_destroy(dev);
		dev_io_destroy(dev);
		dev_irq_destroy(dev);
		gsc_bar_destroy(&dev->bar);
		os_sem_destroy(&dev->sem);
		os_spinlock_destroy(&dev->spinlock);
		os_pci_master_clear(&dev->pci);
		os_pci_dev_disable(&dev->pci);
	}
}


