// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/18AISS8AO8/18AISS8AO8_1.x_GSC_DN/driver/ioctl.c $
// $Rev: 52620 $
// $Date: 2023-03-10 14:56:23 -0600 (Fri, 10 Mar 2023) $

// 18AISS8AO8: Device Driver: source file

#include "main.h"



// macros *********************************************************************

#define	AUTO_CAL_START			D28		// BCR
#define	AUTO_CAL_PASS			D29		// BCR
#define	INIT_START				D31		// BCR

#define	IRQ_AUTO_CAL_DONE		D0		// PSR



//*****************************************************************************
static int _query(dev_data_t* dev, s32* arg)
{
	switch (arg[0])
	{
		default:								arg[0]	= AISS8AO8_IOCTL_QUERY_ERROR;	break;
		case AISS8AO8_QUERY_AI_UNDERFLOW:		arg[0]	= dev->cache.ai_underflow;		break;
		case AISS8AO8_QUERY_AUTO_CAL_MS:		arg[0]	= dev->cache.auto_cal_ms;		break;
		case AISS8AO8_QUERY_CHANNEL_AI_MAX:		arg[0]	= dev->cache.ai_chans_max;		break;
		case AISS8AO8_QUERY_CHANNEL_AI_QTY:		arg[0]	= dev->cache.ai_chan_qty;		break;
		case AISS8AO8_QUERY_CHANNEL_AO_MAX:		arg[0]	= dev->cache.ao_chans_max;		break;
		case AISS8AO8_QUERY_CHANNEL_AO_QTY:		arg[0]	= dev->cache.ao_chan_qty;		break;
		case AISS8AO8_QUERY_COUNT:				arg[0]	= AISS8AO8_IOCTL_QUERY_LAST;	break;
		case AISS8AO8_QUERY_DEVICE_TYPE:		arg[0]	= dev->board_type;				break;
		case AISS8AO8_QUERY_DMDMA:				arg[0]	= dev->cache.dmdma;				break;
		case AISS8AO8_QUERY_FIFO_SIZE_RX:		arg[0]	= dev->cache.fifo_size_rx;		break;
		case AISS8AO8_QUERY_FIFO_SIZE_TX:		arg[0]	= dev->cache.fifo_size_tx;		break;
		case AISS8AO8_QUERY_FILTER_A:			arg[0]	= dev->cache.filter_a;			break;
		case AISS8AO8_QUERY_FILTER_B:			arg[0]	= dev->cache.filter_b;			break;
		case AISS8AO8_QUERY_FSAMP_MAX:			arg[0]	= dev->cache.fsamp_max;			break;
		case AISS8AO8_QUERY_FSAMP_MIN:			arg[0]	= dev->cache.fsamp_min;			break;
		case AISS8AO8_QUERY_INIT_MS:			arg[0]	= dev->cache.initialize_ms;		break;
		case AISS8AO8_QUERY_MASTER_CLOCK:		arg[0]	= dev->cache.master_clock;		break;
		case AISS8AO8_QUERY_NDIV_MAX:			arg[0]	= dev->cache.ndiv_max;			break;
		case AISS8AO8_QUERY_NDIV_MIN:			arg[0]	= dev->cache.ndiv_min;			break;
	}

	return(0);
}



//*****************************************************************************
int initialize_ioctl(dev_data_t* dev, void* arg)
{
	u32		bcr;
	int		i;
	long	ms		= dev->cache.initialize_ms + 5000;
	int		ret		= 0;
	int		tmp;

	// Initiate initialization.
	os_reg_mem_tx_u32(dev, dev->vaddr.gsc_bcr_32, INIT_START);

	// There is no Initialize Done interrupt so we wait manually.
	tmp	= gsc_poll_u32(dev, ms, dev->vaddr.gsc_bcr_32, INIT_START, 0);

	if (tmp)
	{
		ret	= ret ? ret : tmp;
		printf(	"%s: INITIALIZATION DID NOT COMPLETE WITHIN %ld ms.\n",
				dev->model,
				ms);
	}

	// Manually check for completion as the IRQ wait may have ended early.
	bcr	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bcr_32);

	if (bcr & INIT_START)
	{
		ret	= ret ? ret : -ETIMEDOUT;
		printf(	"%s: INITILIZE STILL ACTIVE AFTER %ld ms.\n",
				dev->model,
				ms);
	}

	// Initialize the software settings.

	for (i = 0; i < DEV_IO_STREAM_QTY; i++)
	{
		if (dev->io.io_streams[i])
		{
			if (dev->io.io_streams[i]->dev_io_sw_init)
			{
				(dev->io.io_streams[i]->dev_io_sw_init)(dev, dev->io.io_streams[i]);
			}
		}
	}

	return(ret);
}



//*****************************************************************************
static int _auto_cal_start(dev_data_t* dev, void* arg)
{
	// Initiate auto-calibration.
	os_reg_mem_mx_u32(NULL, dev->vaddr.gsc_bcr_32, AUTO_CAL_START, AUTO_CAL_START);
	return(0);
}



//*****************************************************************************
static int _auto_calibrate(dev_data_t* dev, void* arg)
{
	u32			bcr;
	u32			mask;
	long		ms		= dev->cache.auto_cal_ms + 5000;
	int			ret;
	os_sem_t	sem;
	int			tmp;
	u32			value;
	gsc_wait_t	wt;

	// Safely enable the Auto-Cal Done interrupt.
	mask	= IRQ_AUTO_CAL_DONE;
	value	= IRQ_AUTO_CAL_DONE;
	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_psr_32, value, mask);

	// Wait for the local interrupt.
	os_sem_create(&sem);	// dummy, required for wait operations.
	memset(&wt, 0, sizeof(wt));
	wt.flags		= GSC_WAIT_FLAG_INTERNAL;
	wt.gsc			= AISS8AO8_WAIT_GSC_AUTO_CAL_DONE;
	wt.timeout_ms	= ms;
	ret				= gsc_wait_event(dev, &wt, _auto_cal_start, NULL, &sem);
	os_sem_destroy(&sem);

	if (wt.flags & GSC_WAIT_FLAG_TIMEOUT)
	{
		ret	= ret ? ret : -ETIMEDOUT;
		printf(	"%s: AUTOCALIBRATE IRQ TIMED OUT AFTER %ld ms.\n",
				dev->model,
				ms);
	}

	// Manually wait for completion in case something terminates our wait early.
	tmp	= gsc_poll_u32(dev, ms, dev->vaddr.gsc_bcr_32, AUTO_CAL_START, 0);

	if (tmp)
	{
		ms	*= 2;
		ret	= ret ? ret : tmp;
		printf(	"%s: AUTOCALIBRATION DID NOT COMPLETE WITHIN %ld ms.\n",
				dev->model,
				ms);
	}

	// Manually check for completion as the IRQ wait may have ended early.
	bcr	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bcr_32);

	if (bcr & AUTO_CAL_START)
	{
		ret	= ret ? ret : -ETIMEDOUT;
		printf(	"%s: AUTOCALIBRATION STILL ACTIVE AFTER %ld ms.\n",
				dev->model,
				ms);
	}

	// Final results.
	bcr	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bcr_32);

	if ((bcr & AUTO_CAL_PASS) == 0)
	{
		ret	= ret ? ret : -EIO;
		printf(	"%s: AUTOCALIBRATION FAILED (%ld ms).\n",
				dev->model,
				ms);
	}

	return(ret);
}



//*****************************************************************************
static int _auto_cal_sts(dev_data_t* dev, s32* arg)
{
	u32	reg;
	s32	value;

	reg	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bcr_32);

	if (reg & AUTO_CAL_START)
		value	= AISS8AO8_AUTO_CAL_STS_ACTIVE;
	else if (reg & AUTO_CAL_PASS)
		value	= AISS8AO8_AUTO_CAL_STS_PASS;
	else
		value	= AISS8AO8_AUTO_CAL_STS_FAIL;

	arg[0]	= value;
	return(0);
}



//*****************************************************************************
static int _ai_buf_clear(dev_data_t* dev, void* arg)
{
	int	ret;

	// This action clears the buffer
	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_bcr_32, D13, D13);

	// The bit should clear quickly.
	ret	= gsc_poll_u32(dev, 1, dev->vaddr.gsc_bcr_32, D13, 0);

	// This action clears the overflow flag.
	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_bcr_32, 0, D16);

	// The bit should clear quickly.

	if (ret == 0)
		ret	= gsc_poll_u32(dev, 1, dev->vaddr.gsc_bcr_32, D16, 0);

	return(ret);
}



//*****************************************************************************
static int _ai_buf_level(dev_data_t* dev, s32* arg)
{
	arg[0]	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_ibsr_32);
	arg[0]	&= 0xFFFFF;
	return(0);
}



//*****************************************************************************
static int _ai_buf_overflow(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AISS8AO8_BUF_OVERFLOW_NO,
		AISS8AO8_BUF_OVERFLOW_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 15, 15);
	return(ret);
}



//*****************************************************************************
static int _ai_buf_thr_lvl(dev_data_t* dev, void* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(dev, arg, 0, 0x7FFFF, dev->vaddr.gsc_ibtr_32, 19, 0);
	return(ret);
}



//*****************************************************************************
static int _ai_buf_thr_sts(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AISS8AO8_BUF_THR_STS_CLEAR,
		AISS8AO8_BUF_THR_STS_SET,
		-1	// terminate list
	};

	int	ret;

	arg[0]	= -1;
	ret		= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_ibtr_32, 20, 20);
	return(ret);
}



//*****************************************************************************
static int _ai_burst_enable(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS8AO8_BURST_ENABLE_NO,
		AISS8AO8_BURST_ENABLE_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 9, 9);
	return(ret);
}



//*****************************************************************************
static int _ai_burst_size(dev_data_t* dev, void* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(dev, arg, 0, 0xFFFFFF, dev->vaddr.gsc_ibbsr_32, 23, 0);
	return(ret);
}



//*****************************************************************************
static int _ai_burst_status(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AISS8AO8_AI_BURST_STATUS_BUSY,
		AISS8AO8_AI_BURST_STATUS_READY,
		-1	// terminate list
	};

	int	ret;

	arg[0]	= -1;
	ret		= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 10, 10);
	return(ret);
}



//*****************************************************************************
static int _ai_chan_sel(dev_data_t* dev, void* arg)
{
	int	ret;

	if (dev->cache.ai_chan_qty == 4)
		ret	= gsc_s32_range_reg(dev, arg, 0, 0x0F, dev->vaddr.gsc_icr_32, 7, 0);
	else
		ret	= gsc_s32_range_reg(dev, arg, 0, 0xFF, dev->vaddr.gsc_icr_32, 7, 0);

	return(ret);
}



//*****************************************************************************
static int _ai_clk_trig_out(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS8AO8_AI_CLK_TRIG_OUT_OFF,
		AISS8AO8_AI_CLK_TRIG_OUT_DRIVE,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_icr_32, 11, 11);
	return(ret);
}



//*****************************************************************************
static int _ai_clock_src(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS8AO8_AI_CLOCK_SRC_RAG,
		AISS8AO8_AI_CLOCK_SRC_RCG,
		AISS8AO8_AI_CLOCK_SRC_EXT,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_icr_32, 13, 12);
	return(ret);
}



//*****************************************************************************
static int _ai_enable(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS8AO8_AIO_ENABLE_NO,
		AISS8AO8_AIO_ENABLE_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 12, 12);
	return(ret);
}



//*****************************************************************************
static int _ai_grp_a_filter(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS8AO8_AI_GRP_FILTER_NONE,
		AISS8AO8_AI_GRP_FILTER_A,
		AISS8AO8_AI_GRP_FILTER_B,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_icr_32, 25, 24);
	return(ret);
}



//*****************************************************************************
static int _ai_grp_b_filter(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS8AO8_AI_GRP_FILTER_NONE,
		AISS8AO8_AI_GRP_FILTER_A,
		AISS8AO8_AI_GRP_FILTER_B,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_icr_32, 27, 26);
	return(ret);
}



//*****************************************************************************
static int _ai_i_loop_term(dev_data_t* dev, void* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(dev, arg, 0, 0xFF, dev->vaddr.gsc_clsr_32, 7, 0);
	return(ret);
}



//*****************************************************************************
static int _ai_mode(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS8AO8_AI_MODE_AO_4_HI,
		AISS8AO8_AI_MODE_AO_5_HI,
		AISS8AO8_AI_MODE_AO_6_HI,
		AISS8AO8_AI_MODE_AO_7_HI,
		AISS8AO8_AI_MODE_AO_4_LO,
		AISS8AO8_AI_MODE_AO_5_LO,
		AISS8AO8_AI_MODE_AO_6_LO,
		AISS8AO8_AI_MODE_AO_7_LO,
		AISS8AO8_AI_MODE_DIFF,
		AISS8AO8_AI_MODE_SINGLE,
		AISS8AO8_AI_MODE_ZERO,
		AISS8AO8_AI_MODE_VREF,
		AISS8AO8_AI_MODE_AO_0_HI,
		AISS8AO8_AI_MODE_AO_1_HI,
		AISS8AO8_AI_MODE_AO_2_HI,
		AISS8AO8_AI_MODE_AO_3_HI,
		AISS8AO8_AI_MODE_AO_0_LO,
		AISS8AO8_AI_MODE_AO_1_LO,
		AISS8AO8_AI_MODE_AO_2_LO,
		AISS8AO8_AI_MODE_AO_3_LO,
		-1	// terminate list
	};

	int	ret;

	if (dev->cache.ao_chan_qty == 4)
		ret	= gsc_s32_list_reg(dev, arg, options + 8, dev->vaddr.gsc_icr_32, 20, 16);
	else
		ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_icr_32, 20, 16);

	return(ret);
}



//*****************************************************************************
static int _ai_range(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS8AO8_RANGE_2_5V,
		AISS8AO8_RANGE_5V,
		AISS8AO8_RANGE_10V,
		AISS8AO8_RANGE_0_5V,
		AISS8AO8_RANGE_0_10V,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_icr_32, 10, 8);
	return(ret);
}



//*****************************************************************************
static int _ai_sw_clock(dev_data_t* dev, void* arg)
{
	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_bcr_32, D8, D8);
	return(0);
}



//*****************************************************************************
static int _ai_sw_trigger(dev_data_t* dev, void* arg)
{
	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_bcr_32, D11, D11);
	return(0);
}



//*****************************************************************************
static int _ai_trig_src(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS8AO8_AI_TRIG_SRC_RBG,
		AISS8AO8_AI_TRIG_SRC_RDG,
		AISS8AO8_AI_TRIG_SRC_EXT,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_icr_32, 15, 14);
	return(ret);
}



//*****************************************************************************
static int _ao_buf_clear(dev_data_t* dev, void* arg)
{
	int	ret;

	// This action clears the buffer.
	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_boor_32, D11, D11);

	// The bit should clear quickly.
	ret	= gsc_poll_u32(dev, 1, dev->vaddr.gsc_boor_32, D11, 0);

	// This action clears the buffer overflow and frame overflow bits.
	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_boor_32, 0, D16 | D17);

	// The bit should clear quickly.

	if (ret == 0)
		ret	= gsc_poll_u32(dev, 1, dev->vaddr.gsc_boor_32, D16 | D17, 0);

	return(ret);
}



//*****************************************************************************
static int _ao_buf_empty(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AISS8AO8_AO_BUFFER_EMPTY_NO,
		AISS8AO8_AO_BUFFER_EMPTY_YES,
		-1	// terminate list
	};

	int	ret;

	arg[0]	= -1;
	ret		= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_boor_32, 12, 12);
	return(ret);
}



//*****************************************************************************
static int _ao_buf_full(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS8AO8_AO_BUFFER_FULL_NO,
		AISS8AO8_AO_BUFFER_FULL_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_boor_32, 15, 15);
	return(ret);
}



//*****************************************************************************
static int _ao_buf_level(dev_data_t* dev, s32* arg)
{
	arg[0]	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_obsr_32);
	arg[0]	&= 0xFFFFF;
	return(0);
}



//*****************************************************************************
static int _ao_buf_load_req(dev_data_t* dev, void* arg)
{
	int	ret	= 0;

	// Make the request.
	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_boor_32, D9, D9);

	// Wait for the request to be granted.

	if (dev->io.tx.timeout_s)
		ret	= gsc_poll_u32(dev, dev->io.tx.timeout_s * 1000, dev->vaddr.gsc_boor_32, D10, D10);

	return(ret);
}



//*****************************************************************************
static int _ao_buf_load_sts(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AISS8AO8_BUF_LOAD_STS_NOT_READY,
		AISS8AO8_BUF_LOAD_STS_READY,
		-1	// terminate list
	};

	int	ret;

	arg[0]	= -1;
	ret		= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_boor_32, 10, 10);
	return(ret);
}



//*****************************************************************************
static int _ao_buf_mode(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS8AO8_BUF_MODE_OPEN,
		AISS8AO8_BUF_MODE_CIRC,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_boor_32, 8, 8);
	return(ret);
}



//*****************************************************************************
static int _ao_buf_over_data(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AISS8AO8_BUF_OVERFLOW_NO,
		AISS8AO8_BUF_OVERFLOW_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_boor_32, 16, 16);
	return(ret);
}



//*****************************************************************************
static int _ao_buf_over_fram(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AISS8AO8_BUF_OVERFLOW_NO,
		AISS8AO8_BUF_OVERFLOW_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_boor_32, 17, 17);
	return(ret);
}



//*****************************************************************************
static int _ao_buf_thr_lvl(dev_data_t* dev, void* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(dev, arg, 0, 0x7FFFF, dev->vaddr.gsc_obtr_32, 19, 0);
	return(ret);
}



//*****************************************************************************
static int _ao_buf_thr_sts(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AISS8AO8_BUF_THR_STS_CLEAR,
		AISS8AO8_BUF_THR_STS_SET,
		-1	// terminate list
	};

	int	ret;

	arg[0]	= -1;
	ret		= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_obtr_32, 20, 20);
	return(ret);
}



//*****************************************************************************
static int _ao_burst_enable(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS8AO8_BURST_ENABLE_NO,
		AISS8AO8_BURST_ENABLE_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_boor_32, 19, 19);
	return(ret);
}



//*****************************************************************************
static int _ao_burst_status(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AISS8AO8_AO_BURST_STATUS_BUSY,
		AISS8AO8_AO_BURST_STATUS_READY,
		-1	// terminate list
	};

	int	ret;

	arg[0]	= -1;
	ret		= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_boor_32, 18, 18);
	return(ret);
}



//*****************************************************************************
static int _ao_chan_sel(dev_data_t* dev, void* arg)
{
	int	ret;

	if (dev->cache.ao_chan_qty == 4)
		ret	= gsc_s32_range_reg(dev, arg, 0, 0x0F, dev->vaddr.gsc_ocr_32, 7, 0);
	else
		ret	= gsc_s32_range_reg(dev, arg, 0, 0xFF, dev->vaddr.gsc_ocr_32, 7, 0);

	return(ret);
}



//*****************************************************************************
static int _ao_clk_trig_out(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS8AO8_AO_CLK_TRIG_OUT_OFF,
		AISS8AO8_AO_CLK_TRIG_OUT_DRIVE,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_ocr_32, 11, 11);
	return(ret);
}



//*****************************************************************************
static int _ao_clock_ready(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AISS8AO8_CLOCK_READY_NO,
		AISS8AO8_CLOCK_READY_YES,
		-1	// terminate list
	};

	int	ret;

	arg[0]	= -1;
	ret		= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_boor_32, 6, 6);
	return(ret);
}



//*****************************************************************************
static int _ao_clock_src(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS8AO8_AO_CLOCK_SRC_RCG,
		AISS8AO8_AO_CLOCK_SRC_RAG,
		AISS8AO8_AO_CLOCK_SRC_EXT,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_ocr_32, 13, 12);
	return(ret);
}



//*****************************************************************************
static int _ao_enable(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS8AO8_AIO_ENABLE_NO,
		AISS8AO8_AIO_ENABLE_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_boor_32, 5, 5);
	return(ret);
}



//*****************************************************************************
static int _ao_mode(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS8AO8_AO_MODE_SINGLE,
		AISS8AO8_AO_MODE_DIFF,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_ocr_32, 16, 16);
	return(ret);
}



//*****************************************************************************
static int _ao_output_mode(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS8AO8_AO_OUTPUT_MODE_IMMED,
		AISS8AO8_AO_OUTPUT_MODE_SIMUL,
		-1	// terminate list
	};

	int	ret;

	ret		= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 18, 18);
	return(ret);
}



//*****************************************************************************
static int _ao_range(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS8AO8_RANGE_2_5V,
		AISS8AO8_RANGE_5V,
		AISS8AO8_RANGE_10V,
		AISS8AO8_RANGE_0_5V,
		AISS8AO8_RANGE_0_10V,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_ocr_32, 10, 8);
	return(ret);
}



//*****************************************************************************
static int _ao_sw_clock(dev_data_t* dev, void* arg)
{
	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_boor_32, D7, D7);
	return(0);
}



//*****************************************************************************
static int _ao_sw_trigger(dev_data_t* dev, void* arg)
{
	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_boor_32, D20, D20);
	return(0);
}



//*****************************************************************************
static int _ao_trig_src(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS8AO8_AO_TRIG_SRC_RDG,
		AISS8AO8_AO_TRIG_SRC_RBG,
		AISS8AO8_AO_TRIG_SRC_EXT,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_ocr_32, 15, 14);
	return(ret);
}



//*****************************************************************************
static int _data_format(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS8AO8_DATA_FORMAT_2S_COMP,
		AISS8AO8_DATA_FORMAT_OFF_BIN,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 25, 25);
	return(ret);
}



//*****************************************************************************
static int _dio_dir_out(dev_data_t* dev, s32* arg)
{
	s32	n0;
	s32	n1;
	int	ret	= 0;

	if (arg[0] == -1)
	{
		n0	= -1;
		gsc_s32_range_reg(dev, &n0, 0, 0x1, dev->vaddr.gsc_diopr_32, 7, 7);

		n1	= -1;
		gsc_s32_range_reg(dev, &n1, 0, 0x1, dev->vaddr.gsc_diopr_32, 15, 15);

		arg[0]	= n0 | (n1 << 1);
	}
	else if ((arg[0] & 0x3) == arg[0])
	{
		n0	= (arg[0] >> 0) & 0x1;
		gsc_s32_range_reg(dev, &n0, 0, 0x1, dev->vaddr.gsc_diopr_32, 7, 7);

		n1	= (arg[0] >> 1) & 0x1;
		gsc_s32_range_reg(dev, &n1, 0, 0x1, dev->vaddr.gsc_diopr_32, 15, 15);
	}
	else
	{
		ret	= EINVAL;
	}

	return(ret);
}



//*****************************************************************************
static int _dio_read(dev_data_t* dev, s32* arg)
{
	u32	reg;
	int	ret	= 0;

	reg		= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_diopr_32);
	arg[0]	= ((reg & 0xF00) >> 4) | (reg & 0xF);
	return(ret);
}



//*****************************************************************************
static int _dio_write(dev_data_t* dev, s32* arg)
{
	u32	reg;
	int	ret	= 0;

	if (arg[0] == -1)
	{
		ret	= _dio_read(dev, arg);
	}
	else if ((arg[0] & 0xFFFF) == arg[0])
	{
		reg	= ((arg[0] & 0xF0) << 4) | (arg[0] & 0xF);
		os_reg_mem_mx_u32(dev, dev->vaddr.gsc_diopr_32, reg, 0xF0F);
	}
	else
	{
		ret	= EINVAL;
	}

	return(ret);
}



//*****************************************************************************
static int _gen_a_enable(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS8AO8_GEN_ENABLE_NO,
		AISS8AO8_GEN_ENABLE_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 26, 26);
	return(ret);
}



//*****************************************************************************
static int _gen_a_ndiv(dev_data_t* dev, void* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(
			dev,
			arg,
			dev->cache.ndiv_min,
			dev->cache.ndiv_max,
			dev->vaddr.gsc_ragr_32,
			23, 0);
	return(ret);
}



//*****************************************************************************
static int _gen_b_enable(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS8AO8_GEN_ENABLE_NO,
		AISS8AO8_GEN_ENABLE_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 27, 27);
	return(ret);
}



//*****************************************************************************
static int _gen_b_ndiv(dev_data_t* dev, void* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(
			dev,
			arg,
			dev->cache.ndiv_min,
			dev->cache.ndiv_max,
			dev->vaddr.gsc_rbgr_32,
			23, 0);
	return(ret);
}



//*****************************************************************************
static int _gen_c_enable(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS8AO8_GEN_ENABLE_NO,
		AISS8AO8_GEN_ENABLE_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 21, 21);
	return(ret);
}



//*****************************************************************************
static int _gen_c_ndiv(dev_data_t* dev, void* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(
			dev,
			arg,
			dev->cache.ndiv_min,
			dev->cache.ndiv_max,
			dev->vaddr.gsc_rcgr_32,
			23, 0);
	return(ret);
}



//*****************************************************************************
static int _gen_d_enable(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS8AO8_GEN_ENABLE_NO,
		AISS8AO8_GEN_ENABLE_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 22, 22);
	return(ret);
}




//*****************************************************************************
static int _gen_d_ndiv(dev_data_t* dev, void* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(
			dev,
			arg,
			dev->cache.ndiv_min,
			dev->cache.ndiv_max,
			dev->vaddr.gsc_rdgr_32,
			23, 0);
	return(ret);
}


//*****************************************************************************
static int _irq_enable(dev_data_t* dev, s32* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(dev, arg, 0, AISS8AO8_IRQ_ALL, dev->vaddr.gsc_psr_32, 14, 0);
	return(ret);
}



//*****************************************************************************
static int _rx_io_abort(dev_data_t* dev, s32* arg)
{
	arg[0]	= gsc_read_abort_active_xfer(dev, &dev->io.rx);
	return(0);
}



//*****************************************************************************
static int _rx_io_mode(dev_data_t* dev, s32* arg)
{
	static const s32	list[]	=
	{
		GSC_IO_MODE_DMDMA,
		GSC_IO_MODE_PIO,
		GSC_IO_MODE_BMDMA,
		-1
	};

	int			ret;
	const s32*	lst	= dev->cache.dmdma ? list : (list + 1);

	ret	= gsc_s32_list_var(arg, lst, &dev->io.rx.io_mode);
	return(ret);
}



//*****************************************************************************
static int _rx_io_overflow(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AISS8AO8_IO_OVERFLOW_IGNORE,
		AISS8AO8_IO_OVERFLOW_CHECK,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_var(arg, options, &dev->io.rx.overflow_check);
	return(ret);
}



//*****************************************************************************
static int _rx_io_timeout(dev_data_t* dev, s32* arg)
{
	int	ret;

	ret	= gsc_s32_range_var(
			arg,
			AISS8AO8_IO_TIMEOUT_MIN,
			AISS8AO8_IO_TIMEOUT_INFINITE,
			&dev->io.rx.timeout_s);
	return(ret);
}



//*****************************************************************************
static int _tx_io_abort(dev_data_t* dev, s32* arg)
{
	arg[0]	= gsc_write_abort_active_xfer(dev, &dev->io.tx);
	return(0);
}



//*****************************************************************************
static int _tx_io_mode(dev_data_t* dev, s32* arg)
{
	static const s32	list[]	=
	{
		GSC_IO_MODE_DMDMA,
		GSC_IO_MODE_PIO,
		GSC_IO_MODE_BMDMA,
		-1
	};

	int			ret;
	const s32*	lst	= dev->cache.dmdma ? list : (list + 1);

	ret	= gsc_s32_list_var(arg, lst, &dev->io.tx.io_mode);
	return(ret);
}



//*****************************************************************************
static int _tx_io_overflow(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AISS8AO8_IO_OVERFLOW_IGNORE,
		AISS8AO8_IO_OVERFLOW_CHECK,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_var(arg, options, &dev->io.tx.overflow_check);
	return(ret);
}



//*****************************************************************************
static int _tx_io_timeout(dev_data_t* dev, s32* arg)
{
	int	ret;

	ret	= gsc_s32_range_var(
			arg,
			AISS8AO8_IO_TIMEOUT_MIN,
			AISS8AO8_IO_TIMEOUT_INFINITE,
			&dev->io.tx.timeout_s);
	return(ret);
}



//*****************************************************************************
static int _warp_mode(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS8AO8_WARP_MODE_DISABLE,
		AISS8AO8_WARP_MODE_ENABLE,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 30, 30);
	return(ret);
}



//*****************************************************************************
static int _rx_io_underflow(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AISS8AO8_RX_IO_UNDERFLOW_IGNORE,
		AISS8AO8_RX_IO_UNDERFLOW_CHECK,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_var(arg, options, &dev->io.rx.underflow_check);
	return(ret);
}



//*****************************************************************************
static int _ai_buf_underflow(dev_data_t* dev, s32* arg)
{
	u32	bcr;
	int	ret	= 0;

	switch (arg[0])
	{
		default:

			ret	= -EINVAL;
			break;

		case AISS8AO8_AI_BUF_UNDERFLOW_CLEAR:

			if (dev->cache.ai_underflow)
				os_reg_mem_mx_u32(dev, dev->vaddr.gsc_bcr_32, 0, D16);

			FALLTHROUGH;	/* fall through */

		case AISS8AO8_AI_BUF_UNDERFLOW_TEST:

			if (dev->cache.ai_underflow)
			{
				bcr	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bcr_32);

				if (bcr & D16)
					arg[0]	= AISS8AO8_AI_BUF_UNDERFLOW_YES;
				else
					arg[0]	= AISS8AO8_AI_BUF_UNDERFLOW_NO;
			}
			else
			{
				arg[0]	= -1;
			}

			break;
	}

	return(ret);
}



// variables ******************************************************************

const gsc_ioctl_t	dev_ioctl_list[]	=
{
	{ AISS8AO8_IOCTL_REG_READ,			(void*) gsc_reg_read_ioctl		},
	{ AISS8AO8_IOCTL_REG_WRITE,			(void*) gsc_reg_write_ioctl		},
	{ AISS8AO8_IOCTL_REG_MOD,			(void*) gsc_reg_mod_ioctl		},
	{ AISS8AO8_IOCTL_QUERY,				(void*) _query					},
	{ AISS8AO8_IOCTL_INITIALIZE,		(void*) initialize_ioctl		},
	{ AISS8AO8_IOCTL_AUTO_CALIBRATE,	(void*) _auto_calibrate			},
	{ AISS8AO8_IOCTL_AUTO_CAL_STS,		(void*) _auto_cal_sts			},
	{ AISS8AO8_IOCTL_AI_BUF_CLEAR,		(void*) _ai_buf_clear			},
	{ AISS8AO8_IOCTL_AI_BUF_LEVEL,		(void*) _ai_buf_level			},
	{ AISS8AO8_IOCTL_AI_BUF_OVERFLOW,	(void*) _ai_buf_overflow		},
	{ AISS8AO8_IOCTL_AI_BUF_THR_LVL,	(void*) _ai_buf_thr_lvl			},
	{ AISS8AO8_IOCTL_AI_BUF_THR_STS,	(void*) _ai_buf_thr_sts			},
	{ AISS8AO8_IOCTL_AI_BURST_ENABLE,	(void*) _ai_burst_enable		},
	{ AISS8AO8_IOCTL_AI_BURST_SIZE,		(void*) _ai_burst_size			},
	{ AISS8AO8_IOCTL_AI_BURST_STATUS,	(void*) _ai_burst_status		},
	{ AISS8AO8_IOCTL_AI_CHAN_SEL,		(void*) _ai_chan_sel			},
	{ AISS8AO8_IOCTL_AI_CLK_TRIG_OUT,	(void*) _ai_clk_trig_out		},
	{ AISS8AO8_IOCTL_AI_CLOCK_SRC,		(void*) _ai_clock_src			},
	{ AISS8AO8_IOCTL_AI_ENABLE,			(void*) _ai_enable				},
	{ AISS8AO8_IOCTL_AI_GRP_A_FILTER,	(void*) _ai_grp_a_filter		},
	{ AISS8AO8_IOCTL_AI_GRP_B_FILTER,	(void*) _ai_grp_b_filter		},
	{ AISS8AO8_IOCTL_AI_I_LOOP_TERM,	(void*) _ai_i_loop_term			},
	{ AISS8AO8_IOCTL_AI_MODE,			(void*) _ai_mode				},
	{ AISS8AO8_IOCTL_AI_RANGE,			(void*) _ai_range				},
	{ AISS8AO8_IOCTL_AI_SW_CLOCK,		(void*) _ai_sw_clock			},
	{ AISS8AO8_IOCTL_AI_SW_TRIGGER,		(void*) _ai_sw_trigger			},
	{ AISS8AO8_IOCTL_AI_TRIG_SRC,		(void*) _ai_trig_src			},
	{ AISS8AO8_IOCTL_AO_BUF_CLEAR,		(void*) _ao_buf_clear			},
	{ AISS8AO8_IOCTL_AO_BUF_EMPTY,		(void*) _ao_buf_empty			},
	{ AISS8AO8_IOCTL_AO_BUF_FULL,		(void*) _ao_buf_full			},
	{ AISS8AO8_IOCTL_AO_BUF_LEVEL,		(void*) _ao_buf_level			},
	{ AISS8AO8_IOCTL_AO_BUF_LOAD_REQ,	(void*) _ao_buf_load_req		},
	{ AISS8AO8_IOCTL_AO_BUF_LOAD_STS,	(void*) _ao_buf_load_sts		},
	{ AISS8AO8_IOCTL_AO_BUF_MODE,		(void*) _ao_buf_mode			},
	{ AISS8AO8_IOCTL_AO_BUF_OVER_DATA,	(void*) _ao_buf_over_data		},
	{ AISS8AO8_IOCTL_AO_BUF_OVER_FRAM,	(void*) _ao_buf_over_fram		},
	{ AISS8AO8_IOCTL_AO_BUF_THR_LVL,	(void*) _ao_buf_thr_lvl			},
	{ AISS8AO8_IOCTL_AO_BUF_THR_STS,	(void*) _ao_buf_thr_sts			},
	{ AISS8AO8_IOCTL_AO_BURST_ENABLE,	(void*) _ao_burst_enable		},
	{ AISS8AO8_IOCTL_AO_BURST_STATUS,	(void*) _ao_burst_status		},
	{ AISS8AO8_IOCTL_AO_CHAN_SEL,		(void*) _ao_chan_sel			},
	{ AISS8AO8_IOCTL_AO_CLK_TRIG_OUT,	(void*) _ao_clk_trig_out		},
	{ AISS8AO8_IOCTL_AO_CLOCK_READY,	(void*) _ao_clock_ready			},
	{ AISS8AO8_IOCTL_AO_CLOCK_SRC,		(void*) _ao_clock_src			},
	{ AISS8AO8_IOCTL_AO_ENABLE,			(void*) _ao_enable				},
	{ AISS8AO8_IOCTL_AO_MODE,			(void*) _ao_mode				},
	{ AISS8AO8_IOCTL_AO_OUTPUT_MODE,	(void*) _ao_output_mode			},
	{ AISS8AO8_IOCTL_AO_RANGE,			(void*) _ao_range				},
	{ AISS8AO8_IOCTL_AO_SW_CLOCK,		(void*) _ao_sw_clock			},
	{ AISS8AO8_IOCTL_AO_SW_TRIGGER,		(void*) _ao_sw_trigger			},
	{ AISS8AO8_IOCTL_AO_TRIG_SRC,		(void*) _ao_trig_src			},
	{ AISS8AO8_IOCTL_DATA_FORMAT,		(void*) _data_format			},
	{ AISS8AO8_IOCTL_DIO_DIR_OUT,		(void*) _dio_dir_out			},
	{ AISS8AO8_IOCTL_DIO_READ,			(void*) _dio_read				},
	{ AISS8AO8_IOCTL_DIO_WRITE,			(void*) _dio_write				},
	{ AISS8AO8_IOCTL_GEN_A_ENABLE,		(void*) _gen_a_enable			},
	{ AISS8AO8_IOCTL_GEN_A_NDIV,		(void*) _gen_a_ndiv				},
	{ AISS8AO8_IOCTL_GEN_B_ENABLE,		(void*) _gen_b_enable			},
	{ AISS8AO8_IOCTL_GEN_B_NDIV,		(void*) _gen_b_ndiv				},
	{ AISS8AO8_IOCTL_GEN_C_ENABLE,		(void*) _gen_c_enable			},
	{ AISS8AO8_IOCTL_GEN_C_NDIV,		(void*) _gen_c_ndiv				},
	{ AISS8AO8_IOCTL_GEN_D_ENABLE,		(void*) _gen_d_enable			},
	{ AISS8AO8_IOCTL_GEN_D_NDIV,		(void*) _gen_d_ndiv				},
	{ AISS8AO8_IOCTL_IRQ_ENABLE,		(void*) _irq_enable				},
	{ AISS8AO8_IOCTL_RX_IO_ABORT,		(void*) _rx_io_abort			},
	{ AISS8AO8_IOCTL_RX_IO_MODE,		(void*) _rx_io_mode				},
	{ AISS8AO8_IOCTL_RX_IO_OVERFLOW,	(void*) _rx_io_overflow			},
	{ AISS8AO8_IOCTL_RX_IO_TIMEOUT,		(void*) _rx_io_timeout			},
	{ AISS8AO8_IOCTL_TX_IO_ABORT,		(void*) _tx_io_abort			},
	{ AISS8AO8_IOCTL_TX_IO_MODE,		(void*) _tx_io_mode				},
	{ AISS8AO8_IOCTL_TX_IO_OVERFLOW,	(void*) _tx_io_overflow			},
	{ AISS8AO8_IOCTL_TX_IO_TIMEOUT,		(void*) _tx_io_timeout			},
	{ AISS8AO8_IOCTL_WAIT_CANCEL,		(void*) gsc_wait_cancel_ioctl	},
	{ AISS8AO8_IOCTL_WAIT_EVENT,		(void*) gsc_wait_event_ioctl	},
	{ AISS8AO8_IOCTL_WAIT_STATUS,		(void*) gsc_wait_status_ioctl	},
	{ AISS8AO8_IOCTL_WARP_MODE,			(void*) _warp_mode				},
	{ AISS8AO8_IOCTL_RX_IO_UNDERFLOW,	(void*) _rx_io_underflow		},
	{ AISS8AO8_IOCTL_AI_BUF_UNDERFLOW,	(void*) _ai_buf_underflow		},
	{ -1, NULL }
};


