// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/18AISS8AO8/18AISS8AO8_1.x_GSC_DN/utils/fsamp_ai.c $
// $Rev: 51968 $
// $Date: 2022-10-26 10:38:14 -0500 (Wed, 26 Oct 2022) $

// 18AISS8AO8: Utilities: source file

#include "main.h"



// macros *********************************************************************

#define	ABS(v)					(((v) < 0) ? (-(v)) : (v))
#define	LIMIT_MAX(v,max)		(((v) > (max)) ? (max) : (v))
#define	LIMIT_MIN(v,min)		(((v) < (min)) ? (min) : (v))
#define	LIMIT_RANGE(v,min,max)	(LIMIT_MIN((LIMIT_MAX((v),(max))),(min)))



/******************************************************************************
*
*	Function:	aiss8ao8_fsamp_ai_compute
*
*	Purpose:
*
*		Determine the best configuration for the specified input sample rate.
*
*	Arguments:
*
*		fd		The handle to use to access the driver.
*
*		index	The index of the device to access. Ignore if < 0.
*
*		verbose	Work verbosely?
*
*		fsamp	The desired sample rate. This is quietly adjusted to be valid.
*
*		ndiv	The preferred Ndiv value is put here.
*
*		warp	The preferred warp value is put here.
*
*		rate	Report the sample rate achieved here, if non-NULL.
*
*	Returned:
*
*		>= 0	The number of errors encountered here.
*
******************************************************************************/

int aiss8ao8_fsamp_ai_compute(int fd, int index, int verbose, s32 fsamp, s32* ndiv, s32* warp, double* rate)
{
	double	delta;
	double	delta_best	= DBL_MAX;
	int		errs		= 0;
	double	fsamp_best	= 0;
	double	fsamp_got;
	s32		fsamp_max;		// Queried value.
	s32		fsamp_min;		// Queried value.
	double	fsamp_want;
	s32		master;			// device master clock oscillator frequency
	s32		ndiv_best	= 0;
	s32		ndiv_inc;
	s32		ndiv_max;		// Queried and computet value.
	s32		ndiv_min;		// Queried and computet value.
	s32		warp_best;

	if (verbose)
	{
		gsc_label_index("AI Sample Rate", index);
		gsc_label_long_comma(fsamp);
		printf(" S/S");
	}

	errs	+= aiss8ao8_query(fd, -1, 0, AISS8AO8_QUERY_FSAMP_MAX, &fsamp_max);
	errs	+= aiss8ao8_query(fd, -1, 0, AISS8AO8_QUERY_FSAMP_MIN, &fsamp_min);
	errs	+= aiss8ao8_query(fd, -1, 0, AISS8AO8_QUERY_NDIV_MAX, &ndiv_max);
	errs	+= aiss8ao8_query(fd, -1, 0, AISS8AO8_QUERY_NDIV_MIN, &ndiv_min);
	errs	+= aiss8ao8_query(fd, -1, 0, AISS8AO8_QUERY_MASTER_CLOCK, &master);

	fsamp_want	= (double) LIMIT_RANGE(fsamp, fsamp_min, fsamp_max);

	ndiv_inc	= master / fsamp_want;
	ndiv_max	= LIMIT_RANGE(ndiv_inc + 2, ndiv_min, ndiv_max);
	ndiv_min	= LIMIT_RANGE(ndiv_inc - 2, ndiv_min, ndiv_max);

	for (ndiv_inc = ndiv_min; ndiv_inc <= ndiv_max; ndiv_inc++)
	{
		fsamp_got	= (double) master / ndiv_inc;
		delta		= ABS(fsamp_got - fsamp_want);

		if (delta < delta_best)
		{
			delta_best	= delta;
			fsamp_best	= fsamp_got;
			ndiv_best	= ndiv_inc;
		}
	}

	warp_best	= (fsamp_best > 400000) ? AISS8AO8_WARP_MODE_ENABLE : AISS8AO8_WARP_MODE_DISABLE;

	if (verbose)
	{
		printf("%s  (", errs ? " FAIL <---" : "");
		printf("Fref ");
		gsc_label_long_comma((long) master);
		printf(", Ndiv ");
		gsc_label_long_comma((long) ndiv_best);
		printf(", Warp %s", (warp_best == AISS8AO8_WARP_MODE_ENABLE) ? "Yes" : "No");
		printf(", Fsamp %.3f", fsamp_best);
		printf(")\n");
	}

	if (ndiv)
		ndiv[0]	= ndiv_best;

	if (warp)
		warp[0]	= warp_best;

	if (rate)
		rate[0]	= fsamp_best;

	return(errs);
}



/******************************************************************************
*
*	Function:	aiss8ao8_fsamp_ai_report
*
*	Purpose:
*
*		Determine and report the input sample rate for the specified channel.
*
*	Arguments:
*
*		fd		The handle to use to access the driver.
*
*		index	The index of the device to access. Ignore if < 0.
*
*		verbose	Work verbosely?
*
*		chan	The index of the channel of interest. Ignore if < 0.
*
*		get		Store the sample rate here, if non-NULL.
*
*	Returned:
*
*		>= 0	The number of errors encountered here.
*
******************************************************************************/

int aiss8ao8_fsamp_ai_report(int fd, int index, int verbose, int chan, s32* fsamp)
{
	s32		bit			= 0x1L << chan;
	char	buf[64];
	s32		chans		= 8;
	s32		clk_src;
	int		errs		= 0;
	s32		fsamp_max;
	s32		fsamp_min;
	s32		master;
	s32		ndiv_a;
	s32		ndiv_c;
	s32		rag;
	s32		rate		= 0;
	s32		rcg;
	s32		select;
	s32		warp;

	if (verbose)
	{
		if (chan < 0)
			sprintf(buf, "AI Channels");
		else
			sprintf(buf, "AI Channel %d", chan);

		gsc_label_index(buf, index);
	}

	errs	+= aiss8ao8_query		(fd, -1, 0, AISS8AO8_QUERY_CHANNEL_AI_QTY, &chans);
	errs	+= aiss8ao8_query		(fd, -1, 0, AISS8AO8_QUERY_MASTER_CLOCK, &master);
	errs	+= aiss8ao8_query		(fd, -1, 0, AISS8AO8_QUERY_FSAMP_MAX, &fsamp_max);
	errs	+= aiss8ao8_query		(fd, -1, 0, AISS8AO8_QUERY_FSAMP_MIN, &fsamp_min);
	errs	+= aiss8ao8_ai_chan_sel	(fd, -1, 0, -1, &select);
	errs	+= aiss8ao8_ai_clock_src(fd, -1, 0, -1, &clk_src);
	errs	+= aiss8ao8_gen_a_enable(fd, -1, 0, -1, &rag);
	errs	+= aiss8ao8_gen_a_ndiv	(fd, -1, 0, -1, &ndiv_a);
	errs	+= aiss8ao8_gen_c_enable(fd, -1, 0, -1, &rcg);
	errs	+= aiss8ao8_gen_c_ndiv	(fd, -1, 0, -1, &ndiv_c);
	errs	+= aiss8ao8_warp_mode	(fd, -1, 0, -1, &warp);

	if (errs)
	{
		if (verbose)
			printf("FAIL <---  (query error)\n");
	}
	else if (chan >= 8)
	{
		errs++;

		if (verbose)
			printf("FAIL <---  (Invalid channel index: %ld)\n", (long) chan);
	}
	else if (chan >= chans)
	{
		if (verbose)
			printf("0 S/S  (Channel Not Installed)\n");
	}
	else if ((chan >= 0) && ((select & bit) == 0))
	{
		if (verbose)
			printf("0 S/S  (Channel Disabled)\n");
	}
	else if (clk_src == AISS8AO8_AI_CLOCK_SRC_EXT)
	{
		if (verbose)
			printf("0 S/S  (External Clock Rate Unknown)\n");
	}
	else if (clk_src == AISS8AO8_AI_CLOCK_SRC_RAG)
	{
		if (rag == AISS8AO8_GEN_ENABLE_NO)
		{
			if (verbose)
				printf("0 S/S  (Clock Source Disabled: Rate-A Generator)\n");
		}
		else if (ndiv_a <= 0)
		{
			if (verbose)
				printf("0 S/S  (Invalid Rate-A Generator Ndiv: %ld)\n", (long) ndiv_a);
		}
		else
		{
			rate	= master / ndiv_a;

			if (verbose)
			{
				gsc_label_long_comma(rate);
				printf(" S/S  (Nref ");
				gsc_label_long_comma(master);
				printf(" Hz, RAG, Ndiv ");
				gsc_label_long_comma(ndiv_a);
				printf(", Warp %s", (warp == AISS8AO8_WARP_MODE_ENABLE) ? "Yes" : "No");

				if ((rate < fsamp_min) || (rate > fsamp_max))
					printf(", OUT OF RANGE");

				printf(")\n");
			}
		}
	}
	else
	{
		// clk_src == AISS8AO8_AI_CLOCK_SRC_RCG

		if (rcg == AISS8AO8_GEN_ENABLE_NO)
		{
			if (verbose)
				printf("0 S/S  (Clock Source Disabled: Rate-C Generator)\n");
		}
		else if (ndiv_c <= 0)
		{
			if (verbose)
				printf("0 S/S  (Invalid Rate-C Generator Ndiv: %ld)\n", (long) ndiv_c);
		}
		else
		{
			rate	= master / ndiv_c;

			if (verbose)
			{
				gsc_label_long_comma(rate);
				printf(" S/S  (Nref ");
				gsc_label_long_comma(master);
				printf(" Hz, RCG, Ndiv ");
				gsc_label_long_comma(ndiv_c);
				printf(", Warp %s", (warp == AISS8AO8_WARP_MODE_ENABLE) ? "Yes" : "No");

				if ((rate < fsamp_min) || (rate > fsamp_max))
					printf(", OUT OF RANGE");

				printf(")\n");
			}
		}
	}

	if (fsamp)
		fsamp[0]	= rate;

	return(errs);
}



/******************************************************************************
*
*	Function:	aiss8ao8_fsamp_ai_report_all
*
*	Purpose:
*
*		Determine and report the input sample rate for all channels.
*
*	Arguments:
*
*		fd		The handle to use to access the driver.
*
*		index	The index of the device to access. Ignore if < 0.
*
*		verbose	Work verbosely?
*
*		fsamp	Report overall sample rate here.
*
*	Returned:
*
*		>= 0	The number of errors encountered here.
*
******************************************************************************/

int aiss8ao8_fsamp_ai_report_all(int fd, int index, int verbose, s32* fsamp)
{
	s32	chans;
	int	chan;
	int	errs;
	s32	sps;
	s32	total	= 0;

	if (verbose)
		gsc_label_index("AI Sample Rates", index);

	errs	= aiss8ao8_query(fd, -1, 0, AISS8AO8_QUERY_CHANNEL_AI_QTY, &chans);

	if (errs)
	{
		if (verbose)
			printf("FAIL <---  (channel count query)\n");
	}
	else
	{
		if (verbose)
		{
			printf("\n");
			gsc_label_level_inc();
		}

		for (chan = 0; chan < chans; chan++)
		{
			sps		= 0;
			errs	+= aiss8ao8_fsamp_ai_report(fd, index, verbose, chan, &sps);
			total	+= sps;
		}

		if (verbose)
		{
			gsc_label("Overall AI Rate");
			gsc_label_long_comma(total);
			printf(" S/S\n");
			gsc_label_level_dec();
		}
	}

	if (fsamp)
		fsamp[0]	= total;

	return(errs);
}


